#!/usr/bin/env python
# vim: sw=4

# Canvas items should be able to be constructed like diacanvas.CanvasBox()

import diacanvas as dia
import unittest
import sys, gobject, weakref

class CanvasTestCase(unittest.TestCase):

    def testConstruction (self):
	c = dia.Canvas()
	self.failUnless (sys.getrefcount(c) == 2)
	# One for the wrapper object
	self.failUnless (c.__grefcount__ == 1)
	w_c = weakref.ref(c)
	self.failUnless (isinstance (c, dia.Canvas))
	self.failUnless (isinstance (c.root, dia.CanvasGroup))
	del c
	self.failIf (w_c(), 'Canvas not freed')

    def testAddItem (self):
        c = dia.Canvas()
	self.failUnless (sys.getrefcount(c) == 2)
	self.failUnless (c.__grefcount__ == 1)
	i = dia.CanvasBox()
	self.failUnless (isinstance (i, dia.CanvasBox))
	self.failUnless (sys.getrefcount(i) == 2)
	self.failUnless (i.__grefcount__ == 1)
	w_c = weakref.ref (c)
	w_i = weakref.ref (i)
	i.set (parent=c.root)
	# Only the refcount of the GObject increments ;-)
	self.failUnless (sys.getrefcount(i) == 2)
	self.failUnless (i.__grefcount__ == 2)
	self.failIf (i.parent != c.root)
	self.failIf (c.root.children[0] != i)
	del c, i
	self.failIf (w_c(), 'Canvas not freed')
	self.failIf (w_i(), 'CanvasItem not freed')

    def testRemoveItem (self):
        c = dia.Canvas()
	i = dia.CanvasBox()
	w_c = weakref.ref (c)
	w_i = weakref.ref (i)
	i.set (parent=c.root)
	self.failUnless (isinstance (i, dia.CanvasBox))
	self.failUnless (i.__grefcount__ == 2)
	self.failUnless (len (c.root.children) == 1)
	self.failIf (i.parent != c.root)
	self.failIf (c.root.children[0] != i)
	c.root.remove (i)
	self.failUnless (len (c.root.children) == 0)
	self.failUnless (sys.getrefcount(i) == 2)
	self.failUnless (i.__grefcount__ == 1)
	del c, i
	self.failIf (w_c(), 'Canvas not freed')
	self.failIf (w_i(), 'CanvasItem not freed')

    def testConnection (self):
	c = dia.Canvas()
	b1 = dia.CanvasBox()
	b2 = dia.CanvasBox()
	l = dia.CanvasLine()
	b1.set (parent=c.root)
	self.failUnless (sys.getrefcount(b1) == 2)
	self.failUnless (b1.__grefcount__ == 2)
	c.root.add (b2)
	self.failUnless (sys.getrefcount(b2) == 2)
	self.failUnless (b2.__grefcount__ == 2)
	c.root.add (l)
	self.failUnless (sys.getrefcount(l) == 2)
	self.failUnless (l.__grefcount__ == 2)
	b1.connect_handle(l.handles[0])
	self.failUnless (sys.getrefcount(b1) == 2)
	self.failUnless (b1.__grefcount__ == 2)
	self.failUnless (sys.getrefcount(l) == 2)
	self.failUnless (l.__grefcount__ == 2)
	b2.connect_handle(l.handles[-1])
	self.failUnless (sys.getrefcount(b2) == 2)
	self.failUnless (b2.__grefcount__ == 2)
	self.failUnless (sys.getrefcount(l) == 2)
	self.failUnless (l.__grefcount__ == 2)
	del c, b1, b2, l

    def testConstraint (self):
	c = dia.Canvas()
	con = dia.Constraint()
	v = dia.Variable()
	w_c = weakref.ref (c)
	w_con = weakref.ref (con)
	w_v = weakref.ref (v)
	con.add (v, 1.0)
	c.add_constraint (con)
	del con
	self.failUnless (w_con(), 'Constraint shouldn\'t be freed')
	del v
	self.failUnless (w_v(), 'Variable shouldn\'t be freed')
	del c
	self.failIf (w_c(), 'Canvas not freed')
	self.failIf (w_con(), 'Constraint not freed')
	self.failIf (w_v(), 'Variable not freed')

if __name__ == '__main__':
    unittest.main()
