/* dia-canvas-view-item.c
 * Copyright (C) 2000  Arjan Molenaar
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include "dia-canvas-view.h"
#include "dia-handle-layer.h"
#include "dia-shape.h"
#include "dia-shape-art.h"
#include "dia-shape-x.h"
#include <libart_lgpl/art_uta_svp.h>
#include <libart_lgpl/art_svp_vpath.h>
#include <libart_lgpl/art_svp_ops.h>
#include <libgnomecanvas/gnome-canvas-util.h>
#include "dia-canvas-i18n.h"

//#define D(msg) G_STMT_START { g_print (G_STRLOC": "); g_print msg; g_print ("\n"); } G_STMT_END
#define D(msg)

/* Memory allocation:
 * - In every DiaShape (DiaShapeViewInfo)
 * - handles
 */
/*#define DIA_CANVAS_VIEW_EVENTS	GDK_ALL_EVENTS_MASK
#define DIA_CANVAS_VIEW_EVENTS	GDK_POINTER_MOTION_MASK \
				| GDK_POINTER_MOTION_HINT_MASK \
				| GDK_BUTTON_MOTION_MASK \
				| GDK_BUTTON1_MOTION_MASK \
				| GDK_BUTTON2_MOTION_MASK \
				| GDK_BUTTON3_MOTION_MASK \
				| GDK_BUTTON_PRESS_MASK \
				| GDK_BUTTON_RELEASE_MASK \
				| GDK_KEY_PRESS_MASK \
				| GDK_KEY_RELEASE_MASK */
#define DIA_CANVAS_VIEW_EVENTS (GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_PRESS_MASK)

#define DIA_CANVAS_VIEW_ITEM_USE_GRAB

//#define BB_DEBUG

#ifdef BB_DEBUG
# include <libart_lgpl/art_svp.h>
# include <libart_lgpl/art_uta_rect.h>
# include <libart_lgpl/art_svp_vpath_stroke.h>
# include <libart_lgpl/art_svp_ops.h>
# include <libgnomecanvas/gnome-canvas-util.h>
#endif
enum
{
	PROP_ITEM = 1
};

/* Invariants (constraints that should *always* be true).
 */
#define INVARIANT() \
        g_assert (((DiaCanvasViewItem*)item)->item != NULL); \
        g_assert (DIA_IS_CANVAS_ITEM (((DiaCanvasViewItem*)item)->item));

static void dia_canvas_view_item_init	    (DiaCanvasViewItem *canvas_item);
static void dia_canvas_view_item_class_init (DiaCanvasViewItemClass  *klass);
static void dia_canvas_view_item_set_property	(GObject	*object,
						 guint		 param_id,
						 const GValue	*value,
						 GParamSpec	*pspec);
static void dia_canvas_view_item_get_property	(GObject	*object,
						 guint		 param_id,
						 GValue		*value,
						 GParamSpec	*pspec);
static void dia_canvas_view_item_dispose	(GObject	*object);

static void item_move_cb			(DiaCanvasItem	*item,
						 gdouble dx, gdouble dy,
						 gboolean interactive,
						 DiaCanvasViewItem *vitem);
static void item_need_update_cb			(DiaCanvasItem	*item,
						 GnomeCanvasItem *vitem);
static void item_z_order_cb			(DiaCanvasItem *item,
						 gint positions,
						 DiaCanvasViewItem *vitem);
static void item_state_changed_cb		(DiaCanvasItem *item,
						 gint new_state,
						 DiaCanvasViewItem *vitem);
static gboolean  group_item_add_cb		(DiaCanvasItem	*group,
						 DiaCanvasItem	*item,
						 DiaCanvasViewItem *vitem);
static gboolean  group_item_remove_cb		(DiaCanvasItem	*group,
						 DiaCanvasItem	*item,
						 GnomeCanvasGroup *vgroup);
/* GnomeCanvasItem signals */
static void dia_canvas_view_item_update     (GnomeCanvasItem *item,
					     double *affine,
					     ArtSVP *clip_path, int flags);
static void dia_canvas_view_item_realize    (GnomeCanvasItem *item);
static void dia_canvas_view_item_unrealize  (GnomeCanvasItem *item);
static void dia_canvas_view_item_map        (GnomeCanvasItem *item);
static void dia_canvas_view_item_unmap      (GnomeCanvasItem *item);
static void dia_canvas_view_item_draw       (GnomeCanvasItem *item,
					     GdkDrawable *drawable,
					     int x, int y,
					     int width, int height);
static void dia_canvas_view_item_render     (GnomeCanvasItem *item,
					     GnomeCanvasBuf *buf);
static double dia_canvas_view_item_point    (GnomeCanvasItem *item,
					     double x, double y,
					     int cx, int cy,
					     GnomeCanvasItem **actual_item);
static void dia_canvas_view_item_bounds     (GnomeCanvasItem *item,
					     double *x1, double *y1,
					     double *x2, double *y2);

static gboolean dia_canvas_view_item_event      (GnomeCanvasItem *item,
						 GdkEvent *event);

static void	empty_canvas_view_item		(DiaCanvasViewItem *vitem);

static GnomeCanvasGroupClass *parent_class = NULL;

GtkType
dia_canvas_view_item_get_type (void)
{
	static GtkType canvas_item_type = 0;

	if (!canvas_item_type) {
		static const GtkTypeInfo canvas_item_info =
		{
			"DiaCanvasViewItem",
			sizeof (DiaCanvasViewItem),
			sizeof (DiaCanvasViewItemClass),
			(GtkClassInitFunc) dia_canvas_view_item_class_init,
			(GtkObjectInitFunc) dia_canvas_view_item_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL,
		};
		
		canvas_item_type = gtk_type_unique (gnome_canvas_group_get_type (),
						    &canvas_item_info);
	}

	return canvas_item_type;
}

static void
dia_canvas_view_item_class_init (DiaCanvasViewItemClass *klass)
{
	GObjectClass *object_class;
	GnomeCanvasItemClass *item_class;
  
	object_class = (GObjectClass*) klass;
	item_class = (GnomeCanvasItemClass*) klass;
  
	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = dia_canvas_view_item_dispose;

	object_class->get_property = dia_canvas_view_item_get_property;
	object_class->set_property = dia_canvas_view_item_set_property;

	g_object_class_install_property (object_class,
					 PROP_ITEM,
					 g_param_spec_object ("item",
					 _("Item"),
					 _("DiaCanvasItem that should be visualized"),
					 DIA_TYPE_CANVAS_ITEM,
					 G_PARAM_READABLE
					 | G_PARAM_WRITABLE));

	
	item_class->update = dia_canvas_view_item_update;
	item_class->realize = dia_canvas_view_item_realize;
	item_class->unrealize = dia_canvas_view_item_unrealize;
	item_class->map = dia_canvas_view_item_map;
	item_class->unmap = dia_canvas_view_item_unmap;
	item_class->draw = dia_canvas_view_item_draw;
	item_class->render = dia_canvas_view_item_render;
	item_class->point = dia_canvas_view_item_point;
	item_class->bounds = dia_canvas_view_item_bounds;
	item_class->event = dia_canvas_view_item_event;
}


static void
dia_canvas_view_item_init (DiaCanvasViewItem *item)
{
	item->item = NULL;
	GTK_OBJECT_SET_FLAGS (item, GNOME_CANVAS_ITEM_VISIBLE
			      | GNOME_CANVAS_ITEM_NEED_UPDATE
			      | GNOME_CANVAS_ITEM_NEED_AFFINE
			      | GNOME_CANVAS_ITEM_NEED_VIS
			      | GNOME_CANVAS_ITEM_AFFINE_FULL);
	item->n_handle_pos = 0;
	item->handle_pos = NULL;
	item->event_time = GDK_CURRENT_TIME;
}

static void
dia_canvas_view_item_set_property (GObject *object, guint property_id,
				   const GValue *value, GParamSpec *pspec)
{
	DiaCanvasViewItem *item = DIA_CANVAS_VIEW_ITEM (object);
	
	switch (property_id) {
	case PROP_ITEM:
		if (item->item) {
			g_error ("Setting an item while already an item is set!");
		} else {
			GObject *citem = g_value_get_object (value);
			g_return_if_fail (citem != NULL);
			
			item->item = (DiaCanvasItem*) citem;

			g_signal_connect (G_OBJECT (citem), "move",
					  G_CALLBACK (item_move_cb),
					  item);

			g_signal_connect (G_OBJECT (citem), "need_update",
					  G_CALLBACK (item_need_update_cb),
					  item);

			g_signal_connect (G_OBJECT (citem), "z_order",
					  G_CALLBACK (item_z_order_cb),
					  item);

			g_signal_connect (G_OBJECT (citem), "state_changed",
					  G_CALLBACK (item_state_changed_cb),
					  item);

			/* Make sure the visibility is set correctly: */
			if (DIA_CANVAS_ITEM_FLAGS (citem) & DIA_VISIBLE)
				GTK_OBJECT_SET_FLAGS (item, GNOME_CANVAS_ITEM_VISIBLE);
			else
				GTK_OBJECT_UNSET_FLAGS (item, GNOME_CANVAS_ITEM_VISIBLE);

			/* If our item is a group, notify the CanvasViewItem
			 * that a new object has to be added, if a new object
			 * is added to the DiaCanvas. */
			if (DIA_IS_CANVAS_GROUPABLE (citem)) {
				//g_message (__FUNCTION__": Adding group signals");
				g_signal_connect_after (G_OBJECT (citem),
					"add",
					G_CALLBACK (group_item_add_cb),
					item);
				g_signal_connect_after (G_OBJECT (citem),
					"remove",
					G_CALLBACK (group_item_remove_cb),
					item);
			}

			/* Point directly to the transformation! */
			GNOME_CANVAS_ITEM (item)->xform = ((DiaCanvasItem*) citem)->affine;
			GTK_OBJECT_SET_FLAGS (GTK_OBJECT (item), 
					      GNOME_CANVAS_ITEM_AFFINE_FULL);
			gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (item));
		}
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void
dia_canvas_view_item_get_property (GObject *object, guint property_id,
				   GValue *value, GParamSpec *pspec)
{
	DiaCanvasViewItem *item = DIA_CANVAS_VIEW_ITEM (object);
	
	switch (property_id) {
	case PROP_ITEM:
		g_value_set_object (value, (GObject*) item->item);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void
dia_canvas_view_item_dispose (GObject *object)
{
	DiaCanvasViewItem *vitem = (DiaCanvasViewItem*) object;

	//g_message ("%s: begin (%x)", __FUNCTION__, (int)object);

	if (vitem->item) {
		empty_canvas_view_item (vitem);
	}

	/* Reset the transformation, since it's pointing to the
	 * transformation in DiaCanvasItem::affine! */
	((GnomeCanvasItem*) object)->xform = NULL;

	if (GNOME_CANVAS_ITEM (vitem)->canvas) {
		DiaCanvasView *canvas = DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (vitem)->canvas);
		canvas->selected_items = g_list_remove (canvas->selected_items,
							vitem);

		/* If the object is set as focused object, unset it. */
		if (DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (vitem)->canvas)->focus_item == vitem)
			DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (vitem)->canvas)->focus_item = NULL;
	}

	/* Make sure that removing the item from its parent does not result
	 * in a total freeing of the object. */
	G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
item_move_cb (DiaCanvasItem *item, gdouble dx, gdouble dy,
	      gboolean interactive, DiaCanvasViewItem *vitem)
{
	DiaCanvasView *view = DIA_CANVAS_VIEW (((GnomeCanvasItem* )vitem)->canvas);

	if (interactive && (view == dia_canvas_view_get_active_view ())) {
		gdouble i2w[6];
		/* First transform the dx, dy to world coordinates. */
		dia_canvas_item_affine_i2w (DIA_CANVAS_ITEM (item->parent), i2w);
		i2w[4] = dx * i2w[0] + dy * i2w[2];
		i2w[5] = dx * i2w[1] + dy * i2w[3];
		dia_canvas_view_move (view, i2w[4], i2w[5], vitem);
	}

	GTK_OBJECT_SET_FLAGS (vitem, DIA_CANVAS_VIEW_ITEM_UPDATE_SHAPES);
	gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (vitem));
}

/**
 * item_need_update_cb:
 * @item: 
 * @vitem: 
 * 
 * This function is called for every CanvasViewItem. This way ViewItems are
 * notified that the need updating.
 **/
static void
item_need_update_cb (DiaCanvasItem *item, GnomeCanvasItem *vitem)
{
	g_assert (DIA_IS_CANVAS_ITEM (item));
	g_assert (GNOME_IS_CANVAS_ITEM (vitem));

	gnome_canvas_item_request_update (vitem);
}

static void
item_z_order_cb (DiaCanvasItem *item, gint positions, DiaCanvasViewItem *vitem)
{
	g_assert (DIA_IS_CANVAS_ITEM (item));
	g_assert (GNOME_IS_CANVAS_ITEM (vitem));

	if (positions < 0)
		gnome_canvas_item_lower (GNOME_CANVAS_ITEM (vitem), -positions);
	else if (positions > 0)
		gnome_canvas_item_raise (GNOME_CANVAS_ITEM (vitem), positions);

	/* Both items should be on the same position now. */
	g_assert (dia_canvas_groupable_pos (item->parent, item) == g_list_index (GNOME_CANVAS_GROUP (GNOME_CANVAS_ITEM (vitem)->parent)->item_list, vitem));
}

static void
item_state_changed_cb (DiaCanvasItem *item, gint new_state,
		       DiaCanvasViewItem *vitem)
{
	gint diff;
	DiaCanvasView *view;
	GnomeCanvasItem *gitem = (GnomeCanvasItem*) vitem;
	gboolean need_update = FALSE;

	g_assert (DIA_IS_CANVAS_ITEM (item));
	g_assert (GNOME_IS_CANVAS_ITEM (vitem));

	/* We only want to update the flags that have changed: */
	diff = DIA_CANVAS_ITEM_FLAGS (item) ^ new_state;
	view = (DiaCanvasView*) gitem->canvas;

	/* Visibility has changed! All views are affected. */
	if (diff & DIA_VISIBLE) {
		if (new_state & DIA_VISIBLE) {
			GTK_OBJECT_SET_FLAGS (vitem, GNOME_CANVAS_ITEM_VISIBLE);
		} else {
			GTK_OBJECT_UNSET_FLAGS (vitem, GNOME_CANVAS_ITEM_VISIBLE);
			dia_canvas_view_unselect (view, vitem);
		}
		need_update = TRUE;
	}

	/* The state change only takes effect for the view item of the
	 * active view. */
	if (view == dia_canvas_view_get_active_view ()) {
		if (diff & DIA_SELECT) {
			if (new_state & DIA_SELECT)
				dia_canvas_view_select (view, vitem);
			else
				dia_canvas_view_unselect (view, vitem);
			need_update = TRUE;
		}
		if (diff & DIA_FOCUS) {
			if (new_state & DIA_FOCUS)
				dia_canvas_view_focus (view, vitem);
			else if (dia_canvas_view_item_has_focus (vitem))
				dia_canvas_view_focus (view, NULL);
			need_update = TRUE;
		}


		if (diff & DIA_GRAB) {
			if (new_state & DIA_GRAB)
#ifdef DIA_CANVAS_VIEW_ITEM_USE_GRAB
				gnome_canvas_item_grab (gitem,
							DIA_CANVAS_VIEW_EVENTS,
							NULL,
							vitem->event_time);
#endif
			else
				gnome_canvas_item_ungrab (gitem,
							  vitem->event_time);
			need_update = TRUE;
		}
	}

	if (need_update)
		gnome_canvas_item_request_update ((GnomeCanvasItem*) vitem);
}


/*
 * Signals for DiaCanvasGroups
 */
static gboolean
group_item_add_cb (DiaCanvasItem *group, DiaCanvasItem *item,
		     DiaCanvasViewItem *vitem)
{
	/* So, `item' is added to `group' and now we should create a
	 * new child to `vitem'. */
	g_assert (DIA_IS_CANVAS_GROUPABLE (group));
	g_assert (DIA_IS_CANVAS_ITEM (item));
	g_assert (GNOME_IS_CANVAS_ITEM (vitem));

	//g_message (__FUNCTION__);
	dia_canvas_view_item_add_items (GNOME_CANVAS_GROUP (vitem), item);

	return TRUE;
}

static gboolean
group_item_remove_cb (DiaCanvasItem *group, DiaCanvasItem *item,
		      GnomeCanvasGroup *vgroup)
{
	GList *l;
	DiaCanvasViewItem *vitem = NULL;

	g_assert (DIA_IS_CANVAS_GROUPABLE (group));
	g_assert (DIA_IS_CANVAS_ITEM (item));
	g_assert (GNOME_IS_CANVAS_ITEM (vgroup));

	//g_message (__FUNCTION__);
	for (l = vgroup->item_list; l != NULL; l = l->next) {
		if (DIA_CANVAS_VIEW_ITEM (l->data)->item == item) {
			vitem = l->data;
			break;
		}
	}
	g_assert (vitem != NULL);

	empty_canvas_view_item (vitem);

	g_object_run_dispose (G_OBJECT (vitem));

	return TRUE;
}

static void
empty_canvas_view_item (DiaCanvasViewItem *vitem)
{
	DiaCanvasIter iter;

	if (!vitem->item)
		return;

	g_signal_handlers_disconnect_matched (vitem->item,
					      G_SIGNAL_MATCH_DATA,
					      0, /* signal_id */
					      0, /* detail */
					      NULL, /* closure */
					      NULL, /* func */
					      vitem /* data */);

	/* Remove view info from shapes! */
	if (dia_canvas_item_get_shape_iter (vitem->item, &iter)) do {
		DiaShape *shape = dia_canvas_item_shape_value (vitem->item, &iter);
		DiaShapeViewInfo *view_info;

		view_info = dia_shape_get_view_info (shape, vitem);
		if (view_info) {
			dia_shape_view_info_remove (shape, view_info);
		}
	} while (dia_canvas_item_shape_next (vitem->item, &iter));

	/* Request a redraw of the regions where the handles were, only if the
	 * handle layer is still in business. */
	if (DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (vitem)->canvas)->handle_layer)
		dia_handle_layer_update_handles ((DiaHandleLayer*) DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (vitem)->canvas)->handle_layer, vitem);

	g_free (vitem->handle_pos);
	vitem->n_handle_pos = 0;

	vitem->item = NULL;
}

static guint32
event_get_time (GdkEvent *event)
{
	if (event)
		switch (event->type) {
			case GDK_MOTION_NOTIFY:
			       return event->motion.time;
			case GDK_BUTTON_PRESS:
			case GDK_2BUTTON_PRESS:
			case GDK_3BUTTON_PRESS:
			case GDK_BUTTON_RELEASE:
			       return event->button.time;
			case GDK_KEY_PRESS:
			case GDK_KEY_RELEASE:
			       return event->key.time;
			default:
			       break;
		}
	return GDK_CURRENT_TIME;
}

/* Only copy the states that are view specific to the canvas item. Those
 * flags are: SELECT, FOCUS and GRAB.
 */
static inline void
copy_flags_to_canvas_item (DiaCanvasItem *diaitem, GnomeCanvasItem *item)
{
	/* Make sure the flags represent the state of the CanvasViewItem */

	if (DIA_CANVAS_VIEW_ITEM_GRAB (item))
		DIA_SET_FLAGS (diaitem, DIA_GRAB);
	else
		DIA_UNSET_FLAGS (diaitem, DIA_GRAB);

	if (DIA_CANVAS_VIEW_ITEM_FOCUS (item))
		DIA_SET_FLAGS (diaitem, DIA_FOCUS);
	else
		DIA_UNSET_FLAGS (diaitem, DIA_FOCUS);

	if (DIA_CANVAS_VIEW_ITEM_SELECT (item))
		DIA_SET_FLAGS (diaitem, DIA_SELECT);
	else
		DIA_UNSET_FLAGS (diaitem, DIA_SELECT);
}

/**
 * affine_clip:
 * @clip: 
 * @affine: 
 * @irect: OUT
 *
 * Generate a clip box in canvas coordinates (@irect), given a item
 * relative clip rectangle (@clip) and an affine matrix (@affine).
 *
 * Return value: TRUE if we have a clip path, FALSE if @clip contains no
 * clip path (left == right and top == bottom).
 **/
static inline gboolean
affine_clip (ArtDRect *clip, gdouble *affine, ArtIRect *irect)
{
	ArtDRect drect;

	art_drect_affine_transform (&drect, clip, affine);

	art_drect_to_irect (irect, &drect);

	return (art_irect_empty (irect) ? FALSE : TRUE);
}

/* GnomeCanvasItem signals */
/* clip_path is in canvas coordinates... */
static void
dia_canvas_view_item_update (GnomeCanvasItem *item,
			     double *affine, ArtSVP *clip, int flags)
{
	DiaCanvasItem *diaitem;
	DiaCanvasViewItem *cvitem;
	GnomeCanvas *canvas;
	DiaCanvasIter iter;
	gboolean free_clip = FALSE;
	ArtSVP *new_clip;
	/* Prototype for the update function. */
	ArtSVP* (* update_func) (DiaShape*, DiaCanvasViewItem*, double*, ArtSVP*, int);

	D((" "));

	INVARIANT ();
	
	canvas = item->canvas;
	if (!DIA_CANVAS_VIEW (canvas)->canvas)
		return;

	diaitem = DIA_CANVAS_VIEW_ITEM (item)->item;
	cvitem = DIA_CANVAS_VIEW_ITEM (item);

	g_assert (DIA_IS_CANVAS_ITEM (diaitem));

	//g_message (__FUNCTION__": BoundingBox (%x): %f, %f, %f, %f", 
	//(int)diaitem, item->x1, item->y1, item->x2, item->y2);

	dia_handle_layer_update_handles ((DiaHandleLayer*) DIA_CANVAS_VIEW (canvas)->handle_layer,
					 cvitem);

	/* Update the shapes drawn. */
	if (item->canvas->aa) {
		update_func = dia_shape_art_update;
	} else {
		update_func = dia_shape_x_update;
	}

	if (dia_canvas_item_get_shape_iter (diaitem, &iter)) do {
		DiaShape *shape = dia_canvas_item_shape_value (diaitem, &iter);

		if (shape) {
			/* update render info */
			new_clip = update_func (shape,
						DIA_CANVAS_VIEW_ITEM (item),
						affine, clip, flags);

			if (new_clip && clip && free_clip) {
				art_svp_free (clip);
			}
			if (new_clip) {
				clip = new_clip;
				free_clip = TRUE;
			}
		}
	} while (dia_canvas_item_shape_next (diaitem, &iter));

        if (GNOME_CANVAS_ITEM_CLASS (parent_class)->update)
		GNOME_CANVAS_ITEM_CLASS (parent_class)->update (item, affine,
								clip, flags);

	GTK_OBJECT_UNSET_FLAGS (item, DIA_CANVAS_VIEW_ITEM_UPDATE_SHAPES);

	if (clip && free_clip)
		art_svp_free (clip);

	/* Update the bounding box, taking into account the possible
	 * affine transformation. */
	if (affine) {
		dia_canvas_item_bb_affine (diaitem, affine,
					   &item->x1, &item->y1,
					   &item->x2, &item->y2);
	} else {
		item->x1 = diaitem->bounds.left;
		item->y1 = diaitem->bounds.top;
		item->x2 = diaitem->bounds.right;
		item->y2 = diaitem->bounds.bottom;
	}
}

static void
dia_canvas_view_item_realize (GnomeCanvasItem *item)
{
	INVARIANT ();

	D((" "));

	if (!item->canvas->aa) {
		DIA_CANVAS_VIEW_ITEM (item)->gc = gdk_gc_new (item->canvas->layout.bin_window);
	}

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->realize)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->realize) (item);
}

static void
dia_canvas_view_item_unrealize (GnomeCanvasItem *item)
{
	//INVARIANT ();

	D((" "));

	if (!item->canvas->aa) {
		gdk_gc_unref (DIA_CANVAS_VIEW_ITEM (item)->gc);
		DIA_CANVAS_VIEW_ITEM (item)->gc = NULL;
	}

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->unrealize)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->unrealize) (item);
}

static void
dia_canvas_view_item_map (GnomeCanvasItem *item)
{
	INVARIANT ();

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->map)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->map) (item);
}

static void
dia_canvas_view_item_unmap (GnomeCanvasItem *item)
{
	//INVARIANT ();

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->unmap)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->unmap) (item);
}

static void
dia_canvas_view_item_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
			   int x, int y, int width, int height)
{
	DiaCanvasIter iter;
	DiaCanvasItem *diaitem;

	INVARIANT ();
	
	D((" "));
	diaitem = DIA_CANVAS_VIEW_ITEM (item)->item;

	if (dia_canvas_item_get_shape_iter (diaitem, &iter)) do {
		DiaShape *shape = dia_canvas_item_shape_value (diaitem, &iter);
		DiaShapeVisibility vis = dia_shape_get_visibility (shape);
		if ((vis == DIA_SHAPE_VISIBLE)
		    || ((vis == DIA_SHAPE_VISIBLE_IF_SELECTED)
			&& DIA_CANVAS_VIEW_ITEM_SELECT (item))
		    || ((vis == DIA_SHAPE_VISIBLE_IF_FOCUSED)
			&& DIA_CANVAS_VIEW_ITEM_FOCUS (item))
		    || ((vis == DIA_SHAPE_VISIBLE_IF_GRABBED)
		        && DIA_CANVAS_VIEW_ITEM_GRAB (item)))
			dia_shape_x_render (shape,
					    DIA_CANVAS_VIEW_ITEM (item),
					    drawable, x, y, width, height);
	} while (dia_canvas_item_shape_next (diaitem, &iter));

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->draw)
		GNOME_CANVAS_ITEM_CLASS (parent_class)->draw (item, drawable, x, y, width, height);
}

static void
dia_canvas_view_item_render (GnomeCanvasItem *item, GnomeCanvasBuf *buf)
{
	DiaCanvasIter iter;
	DiaCanvasItem *diaitem;
	D((" "));

	INVARIANT ();
	
	g_assert (DIA_IS_CANVAS_VIEW_ITEM (item));

	diaitem = DIA_CANVAS_VIEW_ITEM (item)->item;
	gnome_canvas_buf_ensure_buf (buf);

	if (!DIA_CANVAS_VIEW (item->canvas)->canvas)
		return;

	if (dia_canvas_item_get_shape_iter (diaitem, &iter)) do {
		DiaShape *shape = dia_canvas_item_shape_value (diaitem, &iter);
		DiaShapeVisibility vis = dia_shape_get_visibility (shape);
		if ((vis == DIA_SHAPE_VISIBLE)
		    || ((vis == DIA_SHAPE_VISIBLE_IF_SELECTED)
			&& DIA_CANVAS_VIEW_ITEM_SELECT (item))
		    || ((vis == DIA_SHAPE_VISIBLE_IF_FOCUSED)
			&& DIA_CANVAS_VIEW_ITEM_FOCUS (item))
		    || ((vis == DIA_SHAPE_VISIBLE_IF_GRABBED)
		        && DIA_CANVAS_VIEW_ITEM_GRAB (item))) {
			dia_shape_art_render (shape,
					      DIA_CANVAS_VIEW_ITEM (item),
					      buf);
		} else {
			// FixMe: clear region of shape anyway...
			// dia_shape_art_clear(shape,
			//		      DIA_CANVAS_VIEW_ITEM (item),
			//		      buf);
		}
	} while (dia_canvas_item_shape_next (diaitem, &iter));
	
#ifdef BB_DEBUG
	G_STMT_START {
		ArtVpath path[6];
		ArtSVP *svp;
		
		path[0].code = ART_MOVETO;
		path[0].x = item->x1;
		path[0].y = item->y1;
		path[1].code = ART_LINETO;
		path[1].x = item->x1;
		path[1].y = item->y2;
		path[2].code = ART_LINETO;
		path[2].x = item->x2;
		path[2].y = item->y2;
		path[3].code = ART_LINETO;
		path[3].x = item->x2;
		path[3].y = item->y1;
		path[4].code = ART_LINETO;
		path[4].x = item->x1;
		path[4].y = item->y1;
		path[5].code = ART_END;
		path[5].x = 0.0;
		path[5].y = 0.0;

		svp = art_svp_vpath_stroke (path, ART_PATH_STROKE_JOIN_MITER,
					    ART_PATH_STROKE_CAP_BUTT, 1.0,
					    4, 0.25);
		gnome_canvas_render_svp (buf, svp, 0x88888888);
		art_free (svp);
	} G_STMT_END;
#endif /* BB_DEBUG */
	//if (GNOME_CANVAS_ITEM_CLASS (parent_class)->render)
	GNOME_CANVAS_ITEM_CLASS (parent_class)->render (item, buf);

	D(("done."));
}

static double
dia_canvas_view_item_point (GnomeCanvasItem *item, double x, double y,
			    int cx, int cy, GnomeCanvasItem **actual_item)
{
	double dist1 = G_MAXDOUBLE, dist2 = G_MAXDOUBLE;
	DiaCanvasItemClass *citem_class;
	GnomeCanvasItem *actual_item1 = NULL, *actual_item2 = NULL;

	INVARIANT ();
	
	D((" "));

	citem_class = DIA_CANVAS_ITEM_GET_CLASS (DIA_CANVAS_VIEW_ITEM (item)->item);
	
	if (!citem_class->point)
		return G_MAXDOUBLE;
	
	if (DIA_CANVAS_ITEM_INTERACTIVE (DIA_CANVAS_VIEW_ITEM (item)->item)) {
		//g_message (__FUNCTION__": interactive object");
		dist1 = citem_class->point (DIA_CANVAS_VIEW_ITEM (item)->item,
					    x, y);
		actual_item1 = item;
	}
	
	dist2 = ((GnomeCanvasItemClass*) parent_class)->point (item, x, y,
							       cx, cy,
							       &actual_item2);

	if (!actual_item2 || dist1 < dist2) {
		*actual_item = actual_item1;
		return dist1;
	}
	*actual_item = actual_item2;
	return dist2;
}

static void
dia_canvas_view_item_bounds (GnomeCanvasItem *item,
			    double *x1, double *y1, double *x2, double *y2)
{
	DiaCanvasItem *citem;

	INVARIANT ();

	D((" "));

	citem = DIA_CANVAS_VIEW_ITEM (item)->item;

	dia_canvas_item_update_now (DIA_CANVAS_VIEW_ITEM (item)->item);
	
	if (item->canvas->need_update)
		gnome_canvas_update_now (item->canvas);

	*x1 = item->x1;
	*y1 = item->y1;
	*x2 = item->x2;
	*y2 = item->y2;
}

static gboolean
dia_canvas_view_item_event (GnomeCanvasItem *item, GdkEvent *event)
{
	static guint event_signal_id = 0;
	DiaCanvasItem *diaitem = NULL;
	DiaCanvasView *view;
	gboolean retval = FALSE;
	DiaEvent ev;

	INVARIANT ();

	view = DIA_CANVAS_VIEW (item->canvas);

	if (!view->canvas)
		return TRUE;

	diaitem = DIA_CANVAS_VIEW_ITEM (item)->item;

	/* Do not allow event handling for non-interactive items: */
	if (!DIA_CANVAS_ITEM_INTERACTIVE (diaitem))
		return FALSE;

	if (event_signal_id == 0)
		event_signal_id = g_signal_lookup ("event",
						   DIA_TYPE_CANVAS_ITEM);

	switch (event->type) {
	case GDK_BUTTON_PRESS:
		if (view->button_press_handled) {
			; /* Do nothing: button press is already handled. */
		} else if ((event->button.state & GDK_CONTROL_MASK)
		    && (DIA_CANVAS_VIEW_ITEM_SELECT (item))) {
			/* Unselect a selected object when CTRL is pressed. */
			dia_canvas_view_unselect (view,
						  DIA_CANVAS_VIEW_ITEM (item));
			dia_canvas_view_focus (view, NULL);
			dia_canvas_item_request_update (diaitem);
			return TRUE;
		} else {
			/* Unselect all entries if no modifiers are pressed and
			 * the object is not already selected. */
			if (!(event->button.state & (GDK_SHIFT_MASK
						     | GDK_CONTROL_MASK))
			    && !DIA_CANVAS_VIEW_ITEM_SELECT (item)) {
				dia_canvas_view_unselect_all (view);
			}
			dia_canvas_view_focus (view, DIA_CANVAS_VIEW_ITEM (item));
			/* Make sure the parent object is not
			 * called, since that might result in
			 * a deselect of the item (again ;-) */
			view->button_press_handled = TRUE;
		}
		/* Fall through */
	case GDK_2BUTTON_PRESS:
	case GDK_3BUTTON_PRESS:
	case GDK_BUTTON_RELEASE:
	case GDK_MOTION_NOTIFY:
	case GDK_KEY_PRESS:
	case GDK_KEY_RELEASE:
		copy_flags_to_canvas_item (diaitem, item);

		dia_canvas_view_gdk_event_to_dia_event (view,
							(DiaCanvasViewItem*) item,
							event, &ev);

		/* Set the event time for grab/ungrab signals. */
		((DiaCanvasViewItem*) item)->event_time = event_get_time (event);
		g_signal_emit (G_OBJECT (diaitem), event_signal_id, 0,
			       &ev, &retval);

		((DiaCanvasViewItem*) item)->event_time = GDK_CURRENT_TIME;

		break;
	default:
		break;
	}

	return retval;
}

/**
 * dia_canvas_view_item_add_items:
 * @item: 
 * @vitem:
 *
 * Add an item to the #DiaCanvasView's tree and connect some signals so it
 * can automatically update itself when something changes
 *
 * @vitem is of type #GnomeCanvasGroup, since the real root item of the
 * #DiaCanvasView is actually a #GnomeCanvasGroup.
 *
 * This function is used internally.
 * 
 * So, @item is added to the canvas and now we should create a new child
 * to @vitem in order to update the #DiaCanvasView (#GnomeCanvas).
 */
void
dia_canvas_view_item_add_items (GnomeCanvasGroup *vitem, DiaCanvasItem *item)
{
	DiaCanvasViewItem *new_vitem;
	gint item_index, vitem_index;
	DiaCanvasIter iter;

	g_assert (DIA_IS_CANVAS_ITEM (item));
	g_assert (GNOME_IS_CANVAS_GROUP (vitem));
	
	new_vitem = (DiaCanvasViewItem*) gnome_canvas_item_new (GNOME_CANVAS_GROUP (vitem), DIA_TYPE_CANVAS_VIEW_ITEM, "item", item, NULL);
	
	g_assert (new_vitem != NULL);
	
	//gnome_canvas_item_request_update (item);

	/* Move the newly created DiaCanvasViewItem to the right place in
	 * the item list (but not for the root item ;-). */
	if (item->parent) {
		item_index = dia_canvas_groupable_pos (item->parent, item);
		vitem_index = g_list_index (vitem->item_list, new_vitem);
		
		/* New items are always added on the top, so if the
		 * vitem's index is higher that item's there's something
		 * very wrong! */
		g_assert (item_index <= vitem_index);
			
		if (item_index < vitem_index) {
			gnome_canvas_item_lower (GNOME_CANVAS_ITEM (new_vitem),
					 vitem_index - item_index);
		}
	}

	/* If item itself has children, add them also */
	if (DIA_IS_CANVAS_GROUPABLE (item)
	    && dia_canvas_groupable_get_iter (item, &iter)) do {
		DiaCanvasItem *val = dia_canvas_groupable_value (item, &iter);

		if (val)
			dia_canvas_view_item_add_items (GNOME_CANVAS_GROUP (new_vitem),
							val);
	} while (dia_canvas_groupable_next (item, &iter));
}

static gint
do_foreach (DiaCanvasViewItem *item,
	    DiaCanvasViewItemForeachFunc func, gpointer data)
{
	GList *l;
	gint result = TRUE;

	if (!func (item, data))
		return FALSE;

	for (l = GNOME_CANVAS_GROUP (item)->item_list; l != NULL; l = l->next)
		if (DIA_IS_CANVAS_VIEW_ITEM (l->data))
			result &= dia_canvas_view_item_foreach ((DiaCanvasViewItem*) l->data,
								 func, data);

	return result;
}

/**
 * dia_canvas_view_item_foreach:
 * @item: 
 * @func: 
 * @data: 
 *
 * Call @func for each canvas view item. If @func returns %FALSE, no more
 * items are called.
 *
 * Return value: 
 **/
gint
dia_canvas_view_item_foreach (DiaCanvasViewItem *item,
			      DiaCanvasViewItemForeachFunc func, gpointer data)
{
	g_return_val_if_fail (DIA_IS_CANVAS_VIEW_ITEM (item), FALSE);
	g_return_val_if_fail (func != NULL, FALSE);

	return do_foreach (item, func, data);
}

/**
 * dia_canvas_view_item_focus:
 * @item: 
 *
 * Check if an item has the focus. If the real focus is held by a child object
 * of @item and the child has the #DIA_COMPOSITE flag, TRUE is returned.
 *
 * Return value: TRUE/FALSE.
 **/
gboolean
dia_canvas_view_item_has_focus (DiaCanvasViewItem *item)
{
	DiaCanvasViewItem *focus_item;

	g_return_val_if_fail (DIA_IS_CANVAS_VIEW_ITEM (item), FALSE);

	focus_item = DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (item)->canvas)->focus_item;
	if (!focus_item)
		return FALSE;

	do {
		if (item == focus_item)
			return TRUE;
		else if (DIA_CANVAS_ITEM_COMPOSITE (focus_item->item))
			focus_item = (DiaCanvasViewItem*) GNOME_CANVAS_ITEM (focus_item)->parent;
		else
			return FALSE;
	} while (focus_item != NULL);

	return FALSE;
}

gboolean
dia_canvas_view_item_is_selected (DiaCanvasViewItem *item)
{
	GList *sel_items;
	
	g_return_val_if_fail (DIA_IS_CANVAS_VIEW_ITEM (item), FALSE);

	sel_items = DIA_CANVAS_VIEW(GNOME_CANVAS_ITEM (item)->canvas)->selected_items;
	
	if (!sel_items)
		return FALSE;

	/* find non composite parent item */
	while (DIA_CANVAS_ITEM_COMPOSITE (item->item))
		item = (DiaCanvasViewItem*) GNOME_CANVAS_ITEM (item)->parent;

	return g_list_find (sel_items, item) ? TRUE : FALSE;

}

